# Relational algebra latex operators

PROJECT_OP = "\\pi"
RENAME_OP = "\\rho"
SELECT_OP = "\\sigma"
ASSIGN_OP = ":="

JOIN_OP = "\\times"
THETA_JOIN_OP = "\\times"
NATURAL_JOIN_OP = "\\bowtie"
FULL_OUTER_JOIN_OP = "\\fullouterjoin"
LEFT_OUTER_JOIN_OP = "\\leftouterjoin"
RIGHT_OUTER_JOIN_OP = "\\rightouterjoin"

DIFFERENCE_OP = "-"
UNION_OP = "\\cup"
INTERSECT_OP = "\\cap"

# Dependency operators
PRIMARY_KEY_OP = "\\text{PK}"
MULTIVALUED_DEPENDENCY_OP = "\\twoheadrightarrow"
FUNCTIONAL_DEPENDENCY_OP = "\\rightarrow"
INCLUSION_EQUIVALENCE_OP = "\\equiv"
INCLUSION_SUBSUMPTION_OP = "\\subseteq"

# Conditional operators
AND_OP = "\\land"
OR_OP = "\\lor"
EQUAL_OP = "\\eq"
NOT_EQUAL_OP = "\\neq"
LESS_THAN_OP = "\\lt"
LESS_THAN_EQUAL_OP = "\\leq"
GREATER_THAN_OP = "\\gt"
GREATER_THAN_EQUAL_OP = "\\geq"
NOT_OP = "\\neg"
DEFINED_OP = "\\text{defined}"
